package com.threed.jpct.benchxl.parts.impl;

import android.util.FloatMath;

import com.threed.jpct.GenericVertexController;
import com.threed.jpct.Object3D;
import com.threed.jpct.SimpleVector;
import com.threed.jpct.World;

/**
 * Based on the orginal source by Mizuki Takase. Optimized for Android by EgonOlsen.
 * 
 * @author Mizuki Takase, EgonOlsen
 * 
 */
public class ClothController extends GenericVertexController {

	private static final long serialVersionUID = 1L;

	private int gridx;
	private int gridy;
	private float spacing;
	private float stff;
	private float mass;
	private float dt;
	private SimpleVector gravity;
	private int numParticles;
	private int numConstraints;
	private int[] index;
	private Particly[] particles;
	private SimpleVector[] forces;
	private SpringConstraint[] constraints;
	private World theWorld;
	private boolean freezeTop;

	/*
	 * when inputting in the grid x and y, don't forget that this is for the
	 * amount of lines that make up the grid.
	 */
	public ClothController(World w, int x, int y, float sp) {
		theWorld = w;
		freezeTop = false;
		gridx = x;
		gridy = y;
		spacing = sp;
		stff = 5f;
		mass = 0.5f; // 0.5
		dt = 0.05f; // 0.05
		gravity = new SimpleVector(0F, 9F, 0F);
	}

	public ClothController(World w, boolean f, int x, int y, float sp) {
		theWorld = w;
		freezeTop = f;
		gridx = x;
		gridy = y;
		spacing = sp;
		stff = 5f;
		mass = 0.5f;
		dt = 0.05f;
		gravity = new SimpleVector(0F, 9F, 0F);
	}

	/*
	 * when inputting in the grid x and y, don't forget that this is for the
	 * amount of lines that make up the grid.
	 */
	public ClothController(World w, boolean f, int x, int y, float sp, float st, float m, float t, SimpleVector g) {
		theWorld = w;
		freezeTop = f;
		gridx = x;
		gridy = y;
		spacing = sp;
		stff = st;
		mass = m;
		dt = t;
		gravity = g;
	}

	public boolean isFreezeTop() {
		return freezeTop;
	}

	public void setFreezeTop(boolean freezeTop) {
		this.freezeTop = freezeTop;
	}
	
	public boolean setup() {
		index = QSort.sort(getSourceMesh());

		numConstraints = (gridx - 1) * gridy + (gridy - 1) * gridx + (gridx - 1) * 2 * (gridy - 1);
		numParticles = gridx * gridy;
		constraints = new SpringConstraint[numConstraints];
		particles = new Particly[numParticles];
		forces = new SimpleVector[numParticles];

		for (int i = 0; i < numParticles; i++) {
			if (i > numParticles - gridx)
				mass = 5;
			particles[i] = new Particly(getSourceMesh()[i].x, getSourceMesh()[i].y, getSourceMesh()[i].z, mass);
			particles[i].velocity = new SimpleVector(0F, 0F, 0F);
		}

		for (int i = 0; i < forces.length; i++) {
			forces[i] = new SimpleVector();
		}

		float diagstff = FloatMath.sqrt(stff * stff + stff * stff);
		float rt2 = FloatMath.sqrt(spacing * spacing + spacing * spacing);
		int off = 0;
		for (int i = 0; i < numParticles; i++) {
			if (i % (gridx) < gridx - 1 && i < numParticles - gridx || i == 0) {
				constraints[off++] = new SpringConstraint(particles[i], particles[i + 1], spacing, stff);
				constraints[off++] = new SpringConstraint(particles[i], particles[i + gridx], spacing, stff);
				constraints[off++] = new SpringConstraint(particles[i], particles[i + gridx + 1], rt2, diagstff);
				if (i % gridx > 0) {
					constraints[off++] = new SpringConstraint(particles[i], particles[i + gridx - 1], rt2, diagstff);
				}
			} else if ((i % (gridx) == gridx - 1) && (i < (gridx * gridy - (gridx)))) {
				constraints[off++] = new SpringConstraint(particles[i], particles[i + gridx], spacing, stff);
				constraints[off++] = new SpringConstraint(particles[i], particles[i + gridx - 1], rt2, diagstff);
			} else if (i > numParticles - 2) {
			} else {
				constraints[off++] = new SpringConstraint(particles[i], particles[i + 1], spacing, stff);
			}
		}
		return true;
	}

	public void apply() {
		resolveConstraints();

		int j;
		if (freezeTop) {
			j = gridx;
		} else {
			j = 0;
		}
		int end = particles.length;
		for (; j < end; j++) {

			Particly particlesj = particles[j];
			SimpleVector forcesj = forces[j];

			particlesj.oldPosition.set(particlesj.currentPosition);

			particles[j].acceleration.set((float) (particlesj.acceleration.x + (forcesj.x / particlesj.mass) - 0.05 * particlesj.velocity.x), (float) (particlesj.acceleration.y
					+ (forcesj.y / particlesj.mass) - 0.05 * particlesj.velocity.y),
					(float) (particlesj.acceleration.z + (forcesj.z / particlesj.mass) - 0.05 * particlesj.velocity.z));

			float vx = particlesj.velocity.x;
			float vy = particlesj.velocity.y;
			float vz = particlesj.velocity.z;
			float p = 1F;

			particlesj.velocity.set(particlesj.velocity.x + particlesj.acceleration.x * dt, particlesj.velocity.y + particlesj.acceleration.y * dt, particlesj.velocity.z
					+ particlesj.acceleration.z * dt);

			particlesj.currentPosition.set(particlesj.currentPosition.x + ((1F - p) * vx + p * particlesj.velocity.x) * dt, particlesj.currentPosition.y
					+ ((1F - p) * vy + p * particlesj.velocity.y) * dt, particlesj.currentPosition.z + ((1F - p) * vz + p * particlesj.velocity.z) * dt);

			particlesj.acceleration.set(0F, 0F, 0F);

			
			if (theWorld.checkCollision(particlesj.currentPosition, particlesj.velocity, 20) != Object3D.NO_OBJECT) {
				particlesj.currentPosition.set(particlesj.oldPosition);
				particlesj.velocity.set(0, 0, 0);
			}

			// apply force
			forcesj.set(gravity);
		}

		for (int i = 0; i < particles.length; i++) {
			getDestinationMesh()[index[i]].set(particles[i].currentPosition);
		}
	}

	void resolveConstraints() {
		for (int j = 0; j < constraints.length; j++) {
			constraints[j].resolveConstraint();
		}
	}

	public Particly getParticle(SimpleVector v) {
		if (v != null) {
			float mindist = 1000.0F*1000.0F;
			int index = 0;
			for (int i = 0; i < particles.length; i++) {
				SimpleVector pic = particles[i].currentPosition;
				float deltax = pic.x - v.x;
				float deltay = pic.y - v.y;
				float distance = /*FloatMath.sqrt(*/deltax * deltax + deltay * deltay/*)*/;

				if (distance < mindist) {
					mindist = distance;
					index = i;
				}
			}
			return particles[index];
		} else
			return null;
	}

}

class SpringConstraint {

	private Particly a, b;
	private float restLen;
	private float stiffness;

	public SpringConstraint(Particly pa, Particly pb, float restlength, float stiff) {
		a = pa;
		b = pb;
		restLen = restlength;
		stiffness = stiff;
	}

	void resolveConstraint() {
		if (a == b)
			return;
		else {
			SimpleVector acurrentPosition = a.currentPosition;
			SimpleVector bcurrentPosition = b.currentPosition;

			SimpleVector aacceleration = a.acceleration;
			SimpleVector bacceleration = b.acceleration;

			SimpleVector avelocity = a.velocity;
			SimpleVector bvelocity = b.velocity;

			float amass1 = 1f / a.mass;
			float bmass1 = 1f / b.mass;

			float deltax = acurrentPosition.x - bcurrentPosition.x;
			float deltay = acurrentPosition.y - bcurrentPosition.y;
			float deltaz = acurrentPosition.z - bcurrentPosition.z;
			float distance = FloatMath.sqrt(deltax * deltax + deltay * deltay + deltaz * deltaz);

			float diff = (distance - restLen) * (stiffness / distance);

			float dvx = avelocity.x - bvelocity.x;
			float dvy = avelocity.y - bvelocity.y;
			float dvz = avelocity.z - bvelocity.z;

			float dmp = dvx * deltax + dvy * deltay + dvz * deltaz;
			dmp /= distance * distance;
			dmp *= 2.0;

			a.acceleration.set((float) (aacceleration.x - deltax * (dmp + diff) * (amass1)), (float) (aacceleration.y - deltay * (dmp + diff) * (amass1)),
					(float) (aacceleration.z - deltaz * (dmp + diff) * (amass1)));

			b.acceleration.set((float) (bacceleration.x + deltax * (dmp + diff) * (bmass1)), (float) (bacceleration.y + deltay * (dmp + diff) * (bmass1)),
					(float) (bacceleration.z + deltaz * (dmp + diff) * (bmass1)));
		}
	}
}

class Particly {
	SimpleVector currentPosition;
	SimpleVector oldPosition;
	SimpleVector acceleration;
	SimpleVector velocity;
	float mass;

	public Particly(float x, float y, float z, float m) {
		this.currentPosition = new SimpleVector(x, y, z);
		this.oldPosition = new SimpleVector(x, y, z);
		this.acceleration = new SimpleVector(0.0001F, 0.0001F, 0.0001F);
		this.velocity = new SimpleVector(0.0001F, 0.0001F, 0.0001F);
		this.mass = m;
	}

	void setPosition(float xnew, float ynew, float znew) {
		this.oldPosition.set(this.currentPosition);
		this.currentPosition.set(xnew, ynew, znew);
	}
}

class QSort {
	static int[] QuickSort(SimpleVector[] a, int[] b, int lo0, int hi0) {
		int lo = lo0;
		int hi = hi0;
		SimpleVector mid;

		if (hi0 > lo0) {
			mid = a[(lo0 + hi0) / 2];
			while (lo <= hi) {
				while ((lo < hi0) && (a[lo].y < mid.y || (a[lo].x < mid.x && a[lo].y == mid.y))) {
					++lo;
				}

				while ((hi > lo0) && (a[hi].y > mid.y || (a[hi].y == mid.y && a[hi].x > mid.x))) {
					--hi;
				}
				if (lo <= hi) {
					swap(a, lo, hi);
					swap(b, lo, hi);

					++lo;
					--hi;
				}
			}
			if (lo0 < hi) {
				QuickSort(a, b, lo0, hi);
			}
			if (lo < hi0) {
				QuickSort(a, b, lo, hi0);
			}
		}

		return b;
	}

	private static final void swap(int a[], int i, int j) {
		int T = a[i];
		a[i] = a[j];
		a[j] = T;
	}

	private static final void swap(SimpleVector a[], int i, int j) {
		SimpleVector T = a[i];
		a[i] = a[j];
		a[j] = T;
	}

	public static final int[] sort(SimpleVector a[]) {
		int[] temp = new int[a.length];
		for (int i = 0; i < temp.length; i++) {
			temp[i] = i;
		}
		QuickSort(a, temp, 0, a.length - 1);
		return temp;
	}
}